/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#include <stdio.h>
#include <string.h> /* for strcmp() */
#include <stdlib.h> /* for sleep() */
#include <pwd.h>    /* for getpwuid() */

/* for time() */
#include <sys/types.h>
#include <sys/time.h>

#include <unistd.h> /* for geteuid() */
#include <ctype.h> /* for isspace() */

#include "burtcommon.h"
#include "burtobjs.h"

#include "list.h"
/*#include "node.h"
  #include "cursor.h"*/
#include "burtlist.h"

#include "SDDS.h"

/****************************************/
/*                                      */
/* Externally Visibile Remote Functions */
/*                                      */
/****************************************/

/* in burtcommon.cc */

void BurtPerformArithmetic(const Burtobjlistlist *const, Burtobjlist *const, 
    const char, const double);
void BurtExit(const int);
BOOL BurtSDDSFile(const char *const);

/* Channel Access Interface Functions */
extern "C" {
void BurtCaFlushIo();
BOOL BurtQueueCaArrayGetCallback(Channel *);
void BurtQueueCaPendEvent(float);
BOOL BurtQueueCaPendIO();
void BurtQueueCaSearchAndConnect(Channel *);
} /* end extern c thing */

/*********************/
/*                   */
/* Utility Functions */
/*                   */
/*********************/

static void setup_sdds_snapshot_table(SDDS_TABLE *, SnapParms *, 
    List<Filename> *const);
static void setup_sdds_request_table(SDDS_TABLE *);

/***************/
/*             */
/* Burtobjlist */
/*             */
/***************/

/**********/
/* Public */
/**********/

/****************************************************************
*
* constructor must be placed here because list.h CANNOT include burtobjs.h.
* it cannont because burtobjs.h needs to include list.h
*
****************************************************************/

Burtobjlist::Burtobjlist()
{
    parms = new SnapParms;
}

/****************************************************************
*
*
****************************************************************/

void Burtobjlist::add_request_filename(const char *const fname)
{

const char *cp;
char cwd[MAXPATHLEN+2];

    if (fname && *fname)
    {
	for (cp = fname; *cp && isspace(*cp); cp ++)
	;

	if (*cp)
	{
	    if (*cp == '/')
		/* absolute pathname */
		get_request_filenames()->append(new Filename(fname));
	    else
	    {
		/* relative pathname */
		if (getcwd(cwd, sizeof(cwd)))
		    get_request_filenames()->append(new Filename(cwd, fname));
		else
		{
		    fprintf(Log_fp,
	    "ERROR: Burtobjlist::add_request_filename() could not getcwd()\n");
		    fflush(Log_fp);
		    BurtExit(1);
		} /* endif */
	    } /* endif */
	} /* endif */
    } /* endif */

}

/****************************************************************
*
* This method receives a pointer (p) to a Burtobj.  The question is
* whether or not this Burtobjlist has an element that refers to the
* same thing that p refers to.  In order for that to be TRUE, the 
* following conditions must be met:
*
* 1) they must both be of the same type (either PVAR or ATOMICDEV)
*
* 2) if (value is defined in both)
*       vals' nelem must be equal in both
*    else
*       nreq for readchannel must be equal in both
*    endif
*
* 3) if (type is PVAR)
*       name (meaning pvname) must be the same in both
*    else
*       name (meaning atomic dev name) must be the same in both
*       lineage must be the same in both
*    endif
*
* NOTE: RO is ignored
*
****************************************************************/

Burtobj *Burtobjlist::find(Burtobj *const p) const
{

Burtobj *rc;
Burtobj *temp;
Node *fnp;

    if (p)
    {
	for (fnp = gethead(), rc = (Burtobj *) NULL; 
	    fnp && !rc; 
		fnp = fnp->getnext())
	{
	    temp = (Burtobj *) fnp->thisone();

	    if (p->isa() == temp->isa() 
		&& !strcmp(p->get_name(), temp->get_name()))
	    {
		/* both the same type and have the same name */

		/* checking lineage the same for atomic devices */
		if (p->isa() != ATOMICDEV || 
		    !strcmp(((Burtatomdev *) p)->get_lineage(), 
			((Burtatomdev *) temp)->get_lineage()))
		{
		    /* if they were both atomic devices, */
		    /* then they both have the same lineage */
		    /* checking both defined => both same nelem */
		    /* else both readchannel same nreq          */
		    if ((p->get_val())->defined() 
			&& (temp->get_val())->defined())
		    {
			if ((p->get_val())->get_nelem() 
			    == (temp->get_val())->get_nelem())
				rc = temp;
		    }
		    else
		    {
			if ((p->get_val())->get_nelem() 
			    == (temp->get_val())->get_nelem())
				rc = temp;
		    } /* endif */
		} /* endif */
	    } /* endif */
	} /* endfor */
    } 
    else
	rc = (Burtobj *) NULL;

    return rc;

}

/****************************************************************
*
*
****************************************************************/

Burtobj *Burtobjlist::find_name(const char *const cp) const
{

Burtobj *rc;
Node *np;

    /* found iff same name (from Pvar) */
    if (cp)
    {
	for (np = gethead(), rc = (Burtobj *) NULL; 
	    np && !rc; 
		np = np->getnext())
	    if (!strcmp(((Burtobj *) (np->thisone()))->get_name(), cp))
		rc = (Burtobj *) np->thisone();
    } 
    else
	rc = (Burtobj *) NULL;

    return rc;

}

/****************************************************************
*
*
****************************************************************/

Burtobj *Burtobjlist::find_name_nreq(const char *const cp, const int nreq) const
{

Burtobj *rc;
Burtobj *temp;
Node *fnp;

    /* found iff same name (from Pvar) and same nreq (from Burtobj) */
    if (cp)
    {
	for (fnp = gethead(), rc = (Burtobj *) NULL; 
	    fnp && !rc; 
		fnp = fnp->getnext())
	{
	    temp = (Burtobj *) fnp->thisone();
	    if (!strcmp((temp->get_readchannel())->get_name(), cp) 
		&& (temp->get_readchannel())->get_nreq() == nreq)
		    rc = temp;
	} /* endfor */
    } 
    else
	rc = (Burtobj *) NULL;

    return rc;

}

/****************************************************************
*
*
****************************************************************/

BOOL Burtobjlist::read_snapshot_from_iocs(const int retry_count)
{
    return read_snapshot_from_iocs(retry_count, (char *const) NULL, 
				    (char *const) NULL);
}

/****************************************************************
*
* class Pvar
*      |
* class Channel
*      |
* class ReadChannel
* 
* this method clears the value found in Pvar and fills it with 
* the value it reads from the ioc.  the cid is in Channel and the
* nreq is in ReadChannel.  the value in Pvar is of type Value
* which has dbr_type, nelem, and pval.  
*
* it is assumed that ca_task_initialize() has been done already, 
* but it is NOT assumed that ca_search()s have been done.
*
****************************************************************/

BOOL Burtobjlist::read_snapshot_from_iocs(const int retry_count, 
    char *const comments, char *const keywords)
{

BOOL rc;

    search_for_read_pvars(retry_count);
    rc = issue_reads();
    rc &= wait_for_reads(retry_count);

    get_parms()->set_keywords((keywords ? keywords : ""));
    get_parms()->set_comments((comments ? comments : ""));
    get_parms()->set_absolute();
    get_parms()->set_who_and_time_stamp();

    return rc;

}

/****************************************************************
*
*
****************************************************************/

void Burtobjlist::write_req_to_file() const
{

Cursor<Burtobj> *rcp;
Burtobj *rp;
SDDS_TABLE SDDS_table;
long row;

    if (Sdds)
    {
	setup_sdds_request_table(&SDDS_table);

	/* starting a new data table */
	if (!SDDS_StartTable(&SDDS_table, (long) nelem()))
	{
	    fprintf(Log_fp, 
"SYS ERR: Burtobjlist::write_req_to_file() SDDS_StartTable().\n");
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	/* writing the values */
	rcp = new Cursor<Burtobj>((List<Burtobj> *const) this);
	for (rp = rcp->getfirst(), row = 0; rp; rp = rcp->getnext(), row ++)
	    rp->write_sdds(&SDDS_table, row);
	delete rcp;

	/* writing the data table */
	if (!SDDS_WriteTable(&SDDS_table))
	{
	    fprintf(Log_fp, 
	    "SYS ERR: Burtobjlist::write_req_to_file() SDDS_Terminate().\n");
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	if (!SDDS_Terminate(&SDDS_table))
	{
	    fprintf(Log_fp, 
	    "SYS ERR: Burtobjlist::write_req_to_file() SDDS_Terminate().\n");
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
    }
    else
    {
	rcp = new Cursor<Burtobj>((List<Burtobj> *const) this);
	for (rp = rcp->getfirst(); rp; rp = rcp->getnext())
	{
	    if (rp->read_only_notify())
		fprintf(Out_fp, "%s ", READONLYNOTIFYSTRING);
	    else if (rp->read_only())
		fprintf(Out_fp, "%s ", READONLYSTRING);

	    fprintf(Out_fp, "%s", (rp->get_readchannel())->get_name());

	    if ((rp->get_readchannel())->get_nreq() != UNKNOWN)
		fprintf(Out_fp, " %d", (rp->get_readchannel())->get_nreq());

	    fprintf(Out_fp, "\n");
	} /* endfor */
	delete rcp;
    } /* endif */

}

/****************************************************************
*
* only writes readchannel values to file, but preserves RO from writechannel
* this is because readchannel RO is never set.
*
****************************************************************/

void Burtobjlist::write_snapshot_to_file() const
{

Cursor<Burtobj> *rcp;
Burtobj *rp;
SDDS_TABLE SDDS_table;
long row;

    if (get_parms()->defined())
    {
	if (Sdds)
	{
	    setup_sdds_snapshot_table(&SDDS_table, get_parms(), 
		look_at_request_filenames());

	    /* starting a new data table */
	    if (!SDDS_StartTable(&SDDS_table, (long) nelem()))
	    {
		fprintf(Log_fp, 
    "SYS ERR: Burtobjlist::write_snapshot_to_file() SDDS_StartTable().\n");
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */

	    /* writing the values */
	    rcp = new Cursor<Burtobj>((List<Burtobj> *const) this);
	    for (rp = rcp->getfirst(), row = 0; 
		rp; rp = rcp->getnext(), row = row + 1)
		    ((rp->get_readchannel())->get_val())->
			write_val_sdds(&SDDS_table, row, rp->get_name(),
					(rp->isa() == PVAR /* type */
					    ? PVSTRING
					    : (rp->isa() == ATOMICDEV
						? DEVSTRING
						: "UNKNOWN")), 
					rp->get_lineage(),
					rp->get_readmsg(),
					rp->get_writemsg(),
					rp->read_only_notify(),
					rp->read_only());
	    delete rcp;

	    /* writing the data table */
	    if (!SDDS_WriteTable(&SDDS_table))
	    {
		fprintf(Log_fp, 
    "SYS ERR: Burtobjlist::write_snapshot_to_file() SDDS_Terminate().\n");
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */

	    if (!SDDS_Terminate(&SDDS_table))
	    {
		fprintf(Log_fp, 
    "SYS ERR: Burtobjlist::write_snapshot_to_file() SDDS_Terminate().\n");
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */
	}
	else
	{
	    write_header();
	    rcp = new Cursor<Burtobj>((List<Burtobj> *const) this);
	    for (rp = rcp->getfirst(); rp; rp = rcp->getnext())
		((rp->get_readchannel())->get_val())->
		    write_val((rp->get_readchannel())->get_name(), 
			rp->read_only_notify(), rp->read_only());
	    delete rcp;
	} /* endif */
    }
    else
    {
	fprintf(Log_fp, 
"ERROR: Burtobjlist::write_snapshot_to_file() attempted to write an undefined snapshot\n");
	BurtExit(1);
    } /* endif */

}

/****************************************************************
*
*
****************************************************************/

BOOL Burtobjlist::write_snapshot_to_iocs(const int retry_count, 
    const char absolute_write_directive, const char relative_write_directive) 
{

BOOL rc;

    if (get_parms()->is_absolute())
    {
	if (absolute_write_directive == ADDWRITE)
	    rc = write_snapshot_to_iocs_relative(retry_count);
	else
	    rc = write_snapshot_to_iocs_absolute(retry_count);
    }
    else if (get_parms()->is_relative())
    {
	if (relative_write_directive == REPLACEWRITE)
	    rc = write_snapshot_to_iocs_absolute(retry_count);
	else
	    rc = write_snapshot_to_iocs_relative(retry_count);
    }
    else if (get_parms()->is_nowrite())
    {
	rc = TRUE;
    }
    else
    {
	fprintf(Log_fp, 
"ERROR: Burtobjlist::write_snapshot_to_iocs() found unrecognizable type\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    return rc;

} 

/***********/
/* Private */
/***********/

/****************************************************************
*
*
****************************************************************/

BOOL Burtobjlist::issue_reads()
{

Cursor<Burtobj> *rcp;
Burtobj *rp;
BOOL issued_read;
BOOL rc;

    if (Verbose)
    {
	fprintf(Log_fp, "\n--- Start reads\n");
	fflush(Log_fp);
    } /* endif */

    rcp = new Cursor<Burtobj>(this);
    for (rp = rcp->getfirst(), issued_read = FALSE, rc = TRUE; 
	    rp; rp = rcp->getnext())
    {
	(rp->get_val())->clear_value();

	(rp->get_readchannel())->clear_read_pending();
	(rp->get_readchannel())->clear_reported();

	if ((rp->get_readchannel())->connected())
	{
	    /* read channel is connected */

	    if ((rp->get_readchannel())->channel_type() == DBR_FLOAT
		     || (rp->get_readchannel())->channel_type() == DBR_CHAR
		     || (rp->get_readchannel())->channel_type() == DBR_DOUBLE)
		/* requesting doubles as doubles and floats as floats */
		/* ... for maximum precision                          */
		(rp->get_readchannel())->
		    set_dbr_req_type((rp->get_readchannel())->channel_type());
	    else
		/* requesting all other types as strings */
		(rp->get_readchannel())-> set_dbr_req_type(DBR_STRING);

	    rc &= BurtQueueCaArrayGetCallback(rp->get_readchannel());
	    issued_read = TRUE;
	}
	else
	{
	    rc = FALSE;

	    fprintf(Log_fp,
		"%s ... not connected so no ca_array_get_callback()\n",
		(rp->get_readchannel())->get_name());
	    fflush(Log_fp);
	} /* endif */
    } /* endfor */
    delete rcp;

    if (Verbose)
    {
	fprintf(Log_fp, "--- End reads\n");
	fflush(Log_fp);
    } /* endif */

    if (issued_read)
	BurtQueueCaPendIO();

    return rc;

}

/****************************************************************
*
*
****************************************************************/

void Burtobjlist::search_for_read_pvars(const int retry_count)
{

Cursor<Burtobj> *rcp;
Burtobj *rp;
unsigned int last_outstanding, nochange_count;
BOOL issued_a_search;

    if (Verbose)
    {
	fprintf(Log_fp, "\n--- Start searches\n");
	fflush(Log_fp);
    } /* endif */

    NoutstandingSearches = 0;

    rcp = new Cursor<Burtobj>(this);
    for (rp = rcp->getfirst(), issued_a_search = FALSE; rp; rp = rcp->getnext())
    {
	if (!((rp->get_readchannel())->connected()))
	{
	    issued_a_search = TRUE;

	    /* increments NoutstandingSearches iff ECA_NORMAL */
	    /* decrements NoutstandingSearches in my_connection_callback() */
	    BurtQueueCaSearchAndConnect((Channel *) rp->get_readchannel());
	} /* endif */
    } /* endfor */
    delete rcp;

    if (Verbose)
    {
	fprintf(Log_fp, "--- End searches\n");
	fflush(Log_fp);
    } /* endif */

    if (issued_a_search)
    {
	last_outstanding = 0;
	nochange_count = 0;
	while (NoutstandingSearches > 0 && nochange_count <= retry_count)
	{

	    if (Verbose)
	    {
		fprintf(Log_fp, 
		    "Waiting for %d outstanding search(es) ...\n", 
		    NoutstandingSearches);
		fflush(Log_fp);
	    } /* endif */

	    BurtQueueCaPendEvent(1.0);

	    if (last_outstanding == NoutstandingSearches)
		nochange_count ++;
	    else
	    {
		last_outstanding = NoutstandingSearches;
		nochange_count = 0;
	    } /* endif */

	} /* endwhile */

	if (Verbose)
	{
	    if (NoutstandingSearches > 0)
		fprintf(Log_fp, "did not find %d\n", NoutstandingSearches);
	    else
		fprintf(Log_fp, "found all OK\n");
	    fflush(Log_fp);
	} /* endif */
    } /* endif */

}

/****************************************************************
*
*
****************************************************************/

void Burtobjlist::search_for_write_pvars(const int retry_count)
{

Cursor<Burtobj> *rcp;
Burtobj *rp;
unsigned int last_outstanding, nochange_count;
BOOL issued_a_search;

    if (Verbose)
    {
	fprintf(Log_fp, "\n--- Start searches\n");
	fflush(Log_fp);
    } /* endif */

    NoutstandingSearches = 0;

    rcp = new Cursor<Burtobj>(this);
    for (rp = rcp->getfirst(), issued_a_search = FALSE; rp; rp = rcp->getnext())
    {
	if (!((rp->get_writechannel())->connected()))
	{
	    issued_a_search = TRUE;

	    /* increments NoutstandingSearches iff ECA_NORMAL */
	    /* decrements NoutstandingSearches in my_connection_callback() */
	    BurtQueueCaSearchAndConnect((Channel *) rp->get_writechannel());
	} /* endif */
    } /* endfor */
    delete rcp;

    if (Verbose)
    {
	fprintf(Log_fp, "--- End searches\n");
	fflush(Log_fp);
    } /* endif */

    if (issued_a_search)
    {
	last_outstanding = 0;
	nochange_count = 0;
	while (NoutstandingSearches > 0 && nochange_count <= retry_count)
	{

	    if (Verbose)
	    {
		fprintf(Log_fp, 
		    "Waiting for %d outstanding search(es) ...\n", 
		    NoutstandingSearches);
		fflush(Log_fp);
	    } /* endif */

	    BurtQueueCaPendEvent(1.0);

	    if (last_outstanding == NoutstandingSearches)
		nochange_count ++;
	    else
	    {
		last_outstanding = NoutstandingSearches;
		nochange_count = 0;
	    } /* endif */

	} /* endwhile */

	if (Verbose)
	{
	    if (NoutstandingSearches > 0)
		fprintf(Log_fp, "did not find %d\n", NoutstandingSearches);
	    else
		fprintf(Log_fp, "found all OK\n");
	    fflush(Log_fp);
	} /* endif */
    } /* endif */

}

/****************************************************************
*
* returns TRUE iff for all channels (read_pending ==> reported)
*
****************************************************************/

BOOL Burtobjlist::wait_for_reads(const int retry_count) const
{

int dotcount;
int minutes;
Cursor<Burtobj> *rcp;
Burtobj *rp;
unsigned int nochange_count, n_outstanding_reads, last_outstanding;
BOOL rc;

    /* waiting for all OK channels to report back */

    if (Verbose)
    {
        fprintf(Log_fp, "\n--- Start wait for pending reads\n");
        fflush(Log_fp);
        dotcount = 0;
        minutes = 0;
    } /* endif */

    rcp = new Cursor<Burtobj>((List<Burtobj> *const) this);

    /* counting reads still outstanding ... read_pending && !reported */
    for (rp = rcp->getfirst(), n_outstanding_reads = 0; rp; rp = rcp->getnext())
	if (rp->get_readchannel() && ((rp->get_readchannel())->read_pending())
	    && !((rp->get_readchannel())->reported()))
		n_outstanding_reads ++;

    last_outstanding = 0;
    nochange_count = 0;
    while (n_outstanding_reads > 0 && nochange_count <= retry_count)
    {
	if (Verbose)
	{
	    fprintf(Log_fp, 
		"Waiting for %d outstanding read(s) ...\n", 
		n_outstanding_reads);
	    fflush(Log_fp);
	} /* endif */

	BurtQueueCaPendEvent(1.0);

	/* counting reads still outstanding ... read_pending && !reported */
	for (rp=rcp->getfirst(), n_outstanding_reads=0; rp; rp = rcp->getnext())
	    if (rp->get_readchannel() 
		&& ((rp->get_readchannel())->read_pending())
		&& !((rp->get_readchannel())->reported()))
		    n_outstanding_reads ++;

	if (last_outstanding == n_outstanding_reads)
	    nochange_count ++;
	else
	{
	    last_outstanding = n_outstanding_reads;
	    nochange_count = 0;
	} /* endif */
    } /* endwhile */
    delete rcp;

    if (Verbose)
    {
        fprintf(Log_fp, 
	    "\n-- End wait for pending reads %d outstanding read(s)\n", 
	    n_outstanding_reads);
        fflush(Log_fp);
    } /* endif */

    rc = (n_outstanding_reads ? FALSE : TRUE);
	
    return rc;

} 

/****************************************************************
*
*
****************************************************************/

BOOL Burtobjlist::write_snapshot_to_iocs_absolute(const int retry_count) 
{

Cursor<Burtobj> *bcp;
Burtobj *bp;
BOOL rc;

    search_for_write_pvars(retry_count);

    if (Verbose)
    {
	fprintf(Log_fp, "\n--- Start writes\n");
	fflush(Log_fp);
    } /* endif */

    bcp = new Cursor<Burtobj>(this);
    for (rc = TRUE, bp = bcp->getfirst(); bp; bp = bcp->getnext())
    {
	if (bp->read_only())
	{
	    if (Verbose)
	    {
		if (bp->read_only_notify())
		{
		    fprintf(Log_fp, 
			">%s< ... read only notify ... no ca_array_put()\n",
			bp->get_name());
		    fflush(Log_fp);
		}
		else
		{
		    fprintf(Log_fp, 
			">%s< ... read only ... no ca_array_put()\n",
			bp->get_name());
		    fflush(Log_fp);
		} /* endif */
	    } /* endif */
	}
	else 
	    rc &= ((bp->get_writechannel())->write_to_ioc());
    } /* endfor */
    delete bcp;

    BurtCaFlushIo();

    if (Verbose)
    {
	fprintf(Log_fp, "--- End writes\n");
	fflush(Log_fp);
    } /* endif */

    return rc;

}

/****************************************************************
*
*
****************************************************************/

BOOL Burtobjlist::write_snapshot_to_iocs_relative(const int retry_count) const
{

Burtobjlist snapshot;
Burtobjlist *netsnapshot;
Burtobjlistlist addends;
Burtpvar *pp;
Burtatomdev *dp;
Cursor<Burtobj> *bcp;
Burtobj *bp;
BOOL rc;

    /* need to take a snapshot of the current values */
    snapshot.set_owner();

    /* loop to populate snapshot ... must look just like this */
    bcp = new Cursor<Burtobj>((List<Burtobj> *const) this);
    for (bp = bcp->getfirst(); bp; bp = bcp->getnext())
    {
	if (bp->isa() == PVAR)
	{
	    if (!(bp->read_only()))
	    {
		pp = new Burtpvar;
		(pp->get_channel())->set_val(pp->get_val());

		(pp->get_channel())->set_name(bp->get_name());
		(pp->get_channel())->set_nreq((bp->get_val())->get_nelem());
		(pp->get_channel())->set_dbr_req_type(NATIVE);

		pp->clear_read_only_notify();
		pp->clear_read_only();

		snapshot.append((Burtobj *) pp);

	    } /* endif */
	}
	else if (bp->isa() == ATOMICDEV)
	{
	    if (!(bp->read_only()))
	    {
		dp = new Burtatomdev;
		(dp->get_readchannel())->set_val(dp->get_val());
		(dp->get_writechannel())->set_val(dp->get_val());


		dp->set_name(bp->get_name());
		dp->set_lineage(bp->get_lineage());

		dp->set_readmsg(bp->get_readmsg());
		(dp->get_readchannel())->set_name((bp->get_readchannel())->
		    get_name());
		(dp->get_readchannel())->
		    set_nreq((bp->get_readchannel())->get_nreq());
		(dp->get_readchannel())->set_dbr_req_type(NATIVE);

		dp->set_writemsg(bp->get_writemsg());
		(dp->get_writechannel())->set_name((bp->get_writechannel())->
		    get_name());
		(dp->get_writechannel())->
		    set_nreq((bp->get_writechannel())->get_nreq());
		(dp->get_writechannel())->set_dbr_req_type(NATIVE);

		dp->clear_read_only_notify();
		dp->clear_read_only();

		snapshot.append((Burtobj *) dp);
	    } /* endif */

	}
	else
	{
	    fprintf(Log_fp, 
"ERROR: Burtobjlist::write_snapshot_to_iocs_relative() found unrecognizable type %d\n", 
		bp->isa());
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
    } /* endfor */
    delete bcp;

    rc = (snapshot.read_snapshot_from_iocs(retry_count));

    addends.append(&snapshot);
    addends.append((const Burtobjlist *const) this);
    netsnapshot = new Burtobjlist;

    BurtPerformArithmetic(&addends, netsnapshot, ADDITION, (double) 1.0);

    (netsnapshot->get_parms())->set_absolute();

    /* must write this one as absolute */
    rc &= (netsnapshot->write_snapshot_to_iocs_absolute(retry_count));

    delete netsnapshot;

    return rc;

}

/****************************************************************
*
*
****************************************************************/

void Burtobjlist::write_header() const
{

Cursor<Filename> *fcp;
Filename *fp;
char cwd[MAXPATHLEN+2];

    fprintf(Out_fp, "%s\n", STARTHEADERSTRING);

    fprintf(Out_fp, "%s %s",
        OLDBURT_TIMEHEADERSTRING, 
    (get_parms()->get_currtime() ? get_parms()->get_currtime() : UNKNOWNTIME));

    /* uid - person who logged in ... effective uid person as */
    /* defined by "set-user-ID" (if done at all)              */
    fprintf(Out_fp, "%s %s (%s)\n",
        OLDBURT_LOGINHEADERSTRING, get_parms()->get_loginid(),
        get_parms()->get_real_world_name());
    fprintf(Out_fp, "%s %d\n",
        OLDBURT_EFFUIDHEADERSTRING, (int) (get_parms()->get_effective_uid()));
    fprintf(Out_fp, "%s %d\n",
        OLDBURT_GROUPIDHEADERSTRING, (int) (get_parms()->get_group_id()));

    fprintf(Out_fp, "%s %s\n",
        OLDBURT_KEYWORDSHEADERSTRING, get_parms()->get_keywords());
    fprintf(Out_fp, "%s %s\n",
        OLDBURT_COMMENTSHEADERSTRING, get_parms()->get_comments());

    if (get_parms()->is_absolute())
        fprintf(Out_fp, "%s %s\n", OLDBURT_TYPEHEADERSTRING, ABSOLUTESTRING);
    else if (get_parms()->is_relative())
        fprintf(Out_fp, "%s %s\n", OLDBURT_TYPEHEADERSTRING, RELATIVESTRING);
    else if (get_parms()->is_nowrite())
        fprintf(Out_fp, "%s %s\n", OLDBURT_TYPEHEADERSTRING, NOWRITESTRING);
    else
    {
        fprintf(Log_fp,
            "ERROR: Burtobjlist::write_header() detected invalid type\n");
        fflush(Log_fp);
        BurtExit(1);
    } /* endif */

    if (look_at_request_filenames()->nelem())
    {
	fcp = new Cursor<Filename>(look_at_request_filenames());
	/* searching for first directory */
	fp = fcp->getfirst();
	while (fp && !(fp->has_dir()))
	    fp = fcp->getnext();

	if (fp && fp->has_dir())
	    /* found at least one with a dir */
	    fprintf(Out_fp, "%s %s\n", OLDBURT_CWDHEADERSTRING, fp->get_dir());
	else
	{
	    /* could not find any with dir ... using cwd */
	    if (getcwd(cwd, sizeof(cwd)))
		fprintf(Out_fp, "%s %s\n", OLDBURT_CWDHEADERSTRING, cwd);
	    else
	    {
		fprintf(Log_fp,
		    "ERROR: Burtobjlist::write_header() could not getcwd()\n");
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */
	} /* endif */

	for (fp = fcp->getfirst(); fp; fp = fcp->getnext())
	    fprintf(Out_fp, "%s %s\n", 
		OLDBURT_REQFNAMEHEADERSTRING, fp->get_fname());
	delete fcp;
    } /* endif */

    fprintf(Out_fp, "%s\n", ENDHEADERSTRING);
    fflush(Out_fp);

}

/*******************/
/*                 */
/* Burtobjlistlist */
/*                 */
/*******************/

Burtobj *Burtobjlistlist::find_name(const char *cp)
{

Burtobj *rc;
Cursor<Burtobjlist> *clcp;
Burtobjlist *clp;

    if (cp)
    {
        clcp = new Cursor<Burtobjlist>(this);
        for (clp = clcp->getfirst(), rc = (Burtobj *) NULL;
            !rc && clp;
                clp = clcp->getnext())
            rc = clp->find_name(cp);
        delete clcp;
    }
    else
        rc = (Burtobj *) NULL;

    return rc;

}

/*********************/
/*                   */
/* Utility Functions */
/*                   */
/*********************/

/****************************************************************
*
*
****************************************************************/

static void setup_sdds_request_table(SDDS_TABLE *table)
{

    if (table)
    {
	if (!SDDS_InitializeOutput(table, SDDS_ASCII, 1L,
		NULL, NULL, Outfilename))
	{
	    fprintf(Log_fp, 
"SYS ERR: setup_sdds_request_table() not SDDS_InitializeOutput() outfile >%s<\n", 
		Outfilename);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	/* defining the columns */
	if (SDDS_DefineColumn(table, NAME_COL, 
	    NULL, NULL, NULL, NULL, SDDS_STRING, 0) == -1)
	{
	    fprintf(Log_fp, 
		"SYS ERR: SDDS_DefineColumn() for col >%s<\n", NAME_COL);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
	if (SDDS_DefineColumn(table, TYPE_COL, 
	    NULL, NULL, NULL, NULL, SDDS_STRING, 0) == -1)
	{
	    fprintf(Log_fp, 
		"SYS ERR: SDDS_DefineColumn() for col >%s<\n", TYPE_COL);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
	if (SDDS_DefineColumn(table, READMSG_COL, 
	    NULL, NULL, NULL, NULL, SDDS_STRING, 0) == -1)
	{
	    fprintf(Log_fp, 
		"SYS ERR: SDDS_DefineColumn() for col >%s<\n", 
		READMSG_COL);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
	if (SDDS_DefineColumn(table, WRITEMSG_COL, 
	    NULL, NULL, NULL, NULL, SDDS_STRING, 0) == -1)
	{
	    fprintf(Log_fp, 
		"SYS ERR: SDDS_DefineColumn() for col >%s<\n", WRITEMSG_COL);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
	if (SDDS_DefineColumn(table, NELEM_COL, 
	    NULL, NULL, NULL, NULL, SDDS_LONG, 0) == -1)
	{
	    fprintf(Log_fp, 
		"SYS ERR: SDDS_DefineColumn() for col >%s<\n", NELEM_COL);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
	if (SDDS_DefineColumn(table, MODE_COL, 
	    NULL, NULL, NULL, NULL, SDDS_STRING, 0) == -1)
	{
	    fprintf(Log_fp, 
		"SYS ERR: SDDS_DefineColumn() for col >%s<\n", MODE_COL);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	/* writing the header */
	if (!SDDS_WriteLayout(table))
	{
	    fprintf(Log_fp, 
		"SYS ERR: setup_sdds_request_table() SDDS_WriteLayout().\n");
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

    }
    else
    {
	fprintf(Log_fp, "SYS ERR: setup_sdds_request_table() got NULL table\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

} /* end setup_sdds_request_table() */

/****************************************************************
*
*
****************************************************************/

static void setup_sdds_snapshot_table(SDDS_TABLE *table, SnapParms *parms,
    List<Filename> *const request_filenames)
{

int nbytes;
char *outbuff;
Cursor<Filename> *fcp;
Filename *fp;

    if (table && parms)
    {
	/* calculating how much we need for outbuff */
	nbytes = 4 + strlen(parms->get_loginid()) 
		+ strlen(parms->get_real_world_name());

	if (nbytes < BYTESFORSTRING(parms->get_effective_uid()))
	    nbytes = BYTESFORSTRING(parms->get_effective_uid());

	if (nbytes < BYTESFORSTRING(parms->get_group_id()))
	    nbytes = BYTESFORSTRING(parms->get_group_id());

	if (!(outbuff = (char *) malloc(nbytes)))
	{
	    fprintf(Log_fp, 
"SYS ERR: setup_sdds_snapshot_table() could not malloc %d bytes for outbuff\n", 
		nbytes);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
		
	if (!SDDS_InitializeOutput(table, SDDS_ASCII, 1L,
		NULL, NULL, Outfilename))
	{
	    fprintf(Log_fp, 
"SYS ERR: setup_sdds_snapshot_table() not SDDS_InitializeOutput() outfile >%s<\n", 
		Outfilename);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	/* defining the columns */
	if (SDDS_DefineColumn(table, NAME_COL, 
	    NULL, NULL, NULL, NULL, SDDS_STRING, 0) == -1)
	{
	    fprintf(Log_fp, "SYS ERR: SDDS_DefineColumn() for col >%s<\n", 
		NAME_COL);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
	if (SDDS_DefineColumn(table, TYPE_COL, 
	    NULL, NULL, NULL, NULL, SDDS_STRING, 0) == -1)
	{
	    fprintf(Log_fp, "SYS ERR: SDDS_DefineColumn() for col >%s<\n", 
		TYPE_COL);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
	if (SDDS_DefineColumn(table, LINEAGE_COL, 
	    NULL, NULL, NULL, NULL, SDDS_STRING, 0) == -1)
	{
	    fprintf(Log_fp, "SYS ERR: SDDS_DefineColumn() for col >%s<\n", 
		LINEAGE_COL);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
	if (SDDS_DefineColumn(table, READMSG_COL, 
	    NULL, NULL, NULL, NULL, SDDS_STRING, 0) == -1)
	{
	    fprintf(Log_fp, "SYS ERR: SDDS_DefineColumn() for col >%s<\n", 
		READMSG_COL);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
	if (SDDS_DefineColumn(table, WRITEMSG_COL, 
	    NULL, NULL, NULL, NULL, SDDS_STRING, 0) == -1)
	{
	    fprintf(Log_fp, "SYS ERR: SDDS_DefineColumn() for col >%s<\n", 
		WRITEMSG_COL);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
	if (SDDS_DefineColumn(table, MODE_COL, 
	    NULL, NULL, NULL, NULL, SDDS_STRING, 0) == -1)
	{
	    fprintf(Log_fp, "SYS ERR: SDDS_DefineColumn() for col >%s<\n", 
		MODE_COL);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
	if (SDDS_DefineColumn(table, NELEM_COL, 
	    NULL, NULL, NULL, NULL, SDDS_LONG, 0) == -1)
	{
	    fprintf(Log_fp, "SYS ERR: SDDS_DefineColumn() for col >%s<\n", 
		NELEM_COL);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
	if (SDDS_DefineColumn(table, VAL_COL, 
	    NULL, NULL, NULL, NULL, SDDS_STRING, 0) == -1)
	{
	    fprintf(Log_fp, "SYS ERR: SDDS_DefineColumn() for col >%s<\n", 
		VAL_COL);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	/* defining the parameters */
	if (!SDDS_DefineParameter(table, TIMEHEADERSTRING, 
	    NULL, NULL, NULL, NULL, SDDS_STRING, 
	(char*)((parms->get_currtime() ? parms->get_currtime() : UNKNOWNTIME))) == -1)
	{
	    fprintf(Log_fp, "SYS ERR: SDDS_DefineParameter() for parm >%s<\n", 
		TIMEHEADERSTRING);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	/* uid - person who logged in ... effective uid person as */
	/* defined by "set-user-ID" (if done at all)              */
	sprintf(outbuff, "%s (%s)", 
	    parms->get_loginid(), parms->get_real_world_name());
	if (!SDDS_DefineParameter(table, LOGINHEADERSTRING, 
	    NULL, NULL, NULL, NULL, SDDS_STRING, outbuff) == -1)
	{
	    fprintf(Log_fp, "SYS ERR: SDDS_DefineParameter() for parm >%s<\n", 
		LOGINHEADERSTRING);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	sprintf(outbuff, "%d", (int) (parms->get_effective_uid()));
	if (!SDDS_DefineParameter(table, EFFUIDHEADERSTRING, 
	    NULL, NULL, NULL, NULL, SDDS_STRING, outbuff) == -1)
	{
	    fprintf(Log_fp, "SYS ERR: SDDS_DefineParameter() for parm >%s<\n", 
		EFFUIDHEADERSTRING);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	sprintf(outbuff, "%d", (int) (parms->get_group_id()));
	if (!SDDS_DefineParameter(table, GROUPIDHEADERSTRING, 
	    NULL, NULL, NULL, NULL, SDDS_STRING, outbuff) == -1)
	{
	    fprintf(Log_fp, "SYS ERR: SDDS_DefineParameter() for parm >%s<\n", 
		GROUPIDHEADERSTRING);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	if (!SDDS_DefineParameter(table, KEYWORDSHEADERSTRING, 
	    NULL, NULL, NULL, NULL, SDDS_STRING, 
	    parms->get_keywords()) == -1)
	{
	    fprintf(Log_fp, "SYS ERR: SDDS_DefineParameter() for parm >%s<\n", 
		KEYWORDSHEADERSTRING);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	if (!SDDS_DefineParameter(table, COMMENTSHEADERSTRING, 
	    NULL, NULL, NULL, NULL, SDDS_STRING, 
	    parms->get_comments()) == -1)
	{
	    fprintf(Log_fp, "SYS ERR: SDDS_DefineParameter() for parm >%s<\n", 
		COMMENTSHEADERSTRING);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	if (parms->is_absolute())
	{
	    if (!SDDS_DefineParameter(table, TYPEHEADERSTRING, 
		NULL, NULL, NULL, NULL, SDDS_STRING, ABSOLUTESTRING) == -1)
	    {
		fprintf(Log_fp, 
		    "SYS ERR: SDDS_DefineParameter() for parm >%s<\n", 
		    TYPEHEADERSTRING);
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */
	}
	else if (parms->is_relative())
	{
	    if (!SDDS_DefineParameter(table, TYPEHEADERSTRING, 
		NULL, NULL, NULL, NULL, SDDS_STRING, RELATIVESTRING) == -1)
	    {
		fprintf(Log_fp, 
		    "SYS ERR: SDDS_DefineParameter() for parm >%s<\n", 
		    TYPEHEADERSTRING);
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */
	}
	else if (parms->is_nowrite())
	{
	    if (!SDDS_DefineParameter(table, TYPEHEADERSTRING, 
		NULL, NULL, NULL, NULL, SDDS_STRING, NOWRITESTRING) == -1)
	    {
		fprintf(Log_fp, 
		    "SYS ERR: SDDS_DefineParameter() for parm >%s<\n", 
		    TYPEHEADERSTRING);
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */
	}
	else
	{
	    fprintf(Log_fp, 
		"ERROR: setup_sdds_snapshot_table() detected invalid type\n");
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	/* defining the associates (req files) */
	fcp = new Cursor<Filename>(request_filenames);
	for (fp = fcp->getfirst(); fp; fp = fcp->getnext())
	{
	    if (SDDS_DefineAssociate(table, "burtSnapshot", (char *) fp->get_fname(), 
		    (char *) fp->get_dir(), (char *) NULL, (char *) NULL, 
		    BurtSDDSFile(fp->get_fname())) == -1)
	    {
		/* something wrong */
		fprintf(Log_fp,
"SYS ERR: setup_sdds_snapshot_table() could not SDDS_DefineAssociate() for req file >%s<\n", 
		    fp->get_fname());
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */
	} /* endfor */
	delete fcp;

	/* writing the header */
	if (!SDDS_WriteLayout(table))
	{
	    fprintf(Log_fp, 
		"SYS ERR: setup_sdds_snapshot_table() SDDS_WriteLayout().\n");
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	free(outbuff);

    }
    else
    {
	fprintf(Log_fp, 
	    "SYS ERR: setup_sdds_snapshot_table() got table %x parms %x\n", 
	    table, parms);
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

} /* end setup_sdds_snapshot_table() */

