/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#define MAINLINE

#include <stdio.h>
#include <stdlib.h>

#include "burtcommon.h"
#include "burtobjs.h"
#include "burtmath.h"

#include "list.h"
#include "burtlist.h"

/****************************************/
/*                                      */
/* Externally Visibile Remote Functions */
/*                                      */
/****************************************/

/* in burtcommon.cc */
void BurtCommonInit(const BOOL);
void BurtPerformArithmetic(const Burtobjlistlist *const, Burtobjlist *const, 
    const char, const double);
Burtobjlist *BurtProcessSnapshotFile(char *const);
BOOL BurtSDDSFile(const char *const);
void BurtExit(const int);

/***************************/
/*                         */
/* Private Local Functions */
/*                         */
/***************************/

static void init();
static void parse_command_line_args(int, char **, char **, char **, char *,
    char ***, int *, char **, char **, BOOL *, char *, double *);
static void process_command_line_args(char *, char *, char **, int,
    Burtobjlistlist *, char);

/****************************************************************
*
*
****************************************************************/

int main(int argc, char **argv)
{

double mult_fact;
char *logfilename;
char **snapfilenames;
int nsnapsfilenames;
char *comments;
char *keywords;
char operation;
Burtobjlistlist snapshots;
Burtobjlist result_snapshot;
char output_spec;

    init();

    parse_command_line_args(argc, argv, &logfilename, &Outfilename, &operation,
        &snapfilenames, &nsnapsfilenames, &comments, &keywords, &Verbose, 
	&output_spec, &mult_fact);

    process_command_line_args(logfilename, Outfilename, snapfilenames,
	nsnapsfilenames, &snapshots, output_spec);

    if (snapfilenames)
	free((char *) snapfilenames);

    BurtPerformArithmetic(&snapshots, &result_snapshot, operation, mult_fact);

    if (keywords)
    {
	(result_snapshot.get_parms())->set_keywords(keywords);
	free(keywords);
    } /* endif */
    else
	/* needed to define the snapshot */
	(result_snapshot.get_parms())->set_keywords("");

    if (comments)
    {
	(result_snapshot.get_parms())->set_comments(comments);
	free(comments);
    } /* endif */
    else
	/* needed to define the snapshot */
	(result_snapshot.get_parms())->set_comments("");

    (result_snapshot.get_parms())->set_who_and_time_stamp();

    result_snapshot.write_snapshot_to_file();

    BurtExit(0);

} /* end main() */

/******************************************************
*
*
******************************************************/

static void init()
{
    BurtCommonInit(FALSE);
} /* end init() */

/******************************************************
*
*
******************************************************/

static void parse_command_line_args(int argc, char **argv, char **logfilename, 
    char **outfilename, char *operation, char ***snapfilenames, 
    int *nsnapfilenames, char **comments, char **keywords, BOOL *verbose, 
    char *output_spec, double *mult_factor)
{

int arg;
int context;
int nbytes;
int i;
char *ret_cp;
BOOL got_all_comments;
BOOL got_all_keywords;
BOOL gotseconds;
BOOL gotoperation;
BOOL gotmultfact;
BOOL error;

    /* default values */
    *logfilename = (char *) NULL;
    *outfilename = (char *) NULL;
    *operation = SUBTRACTION;
    *snapfilenames = (char **) NULL;
    *nsnapfilenames = 0;
    *comments = (char *) NULL;
    *keywords = (char *) NULL;
    *verbose = FALSE;
    *output_spec = NOREQ;
    *mult_factor = 1.0;

    context = SNAP_FILES;
    got_all_comments = FALSE;
    got_all_keywords = FALSE;
    gotseconds = FALSE;
    gotoperation = FALSE;
    gotmultfact = FALSE;
    arg = 1;
    error = FALSE;
    while (arg < argc && !error)
    {
/* printf("error %c argc %d arg %d >%s< context %s\n", (error ? 'T' : 'F'), argc, arg, argv[arg], (context == EXPECT_SWITCH ? "expect switch" : (context == SNAP_FILES ? "snap files" : (context == LOG_FILE ? "log file" : "unknown")))); */
	switch (context)
	{
	    case EXPECT_SWITCH:
		if (strcmp(argv[arg], "-l") == 0)
		{
		    if (*logfilename)
			/* already have a logfile from */
			/* a previous -l switch        */
			error = TRUE;
		    else
		    {
			arg ++;
			context = LOG_FILE;
		    } /* endif */
		}
                else if (strcmp(argv[arg], "-o") == 0)
                {
                    if (*outfilename)
                        /* already have a outfile from */
                        /* a previous -o switch        */
                        error = TRUE;
                    else
                    {
                        arg ++;
                        context = OUT_FILE;
                    } /* endif */
                }
                else if (strcmp(argv[arg], "-c") == 0)
                {
                    if (got_all_comments)
                        /* already got all the comments from */
                        /* from a previous -c switch         */
                        error = TRUE;
                    else
                    {
                        arg ++;
                        context = COMMENTS;
                    } /* endif */
                }
                else if (strcmp(argv[arg], "-k") == 0)
                {
                    if (got_all_keywords)
                        /* already got all the keywords from */
                        /* from a previous -k switch         */
                        error = TRUE;
                    else
                    {
                        arg ++;
                        context = KEYWORDS;
                    } /* endif */
                }
		else if (strcmp(argv[arg], "-v") == 0)
		{
		    arg ++;
		    *verbose = TRUE;
		}
                else if (strcmp(argv[arg], "-sdds") == 0)
                {
                    if (*output_spec == NOREQ)
                    {
                        arg ++;
                        *output_spec = REQSDDS;
                    }
                    else
                        error = TRUE;
                }
                else if (strcmp(argv[arg], "-nosdds") == 0)
                {
                    if (*output_spec == NOREQ)
                    {
                        arg ++;
                        *output_spec = REQNOSDDS;
                    }
                    else
                        error = TRUE;
                }
		else if (strcmp(argv[arg], "-add") == 0)
		{
		    if (gotoperation)
                        /* already have a operation from  */
                        /* a previous -add or -sub switch */
			error = TRUE;
		    else
		    {
			arg ++;
			*operation = ADDITION;
		    } /* endif */
		}
		else if (strcmp(argv[arg], "-sub") == 0)
		{
		    if (gotoperation)
                        /* already have a operation from  */
                        /* a previous -add or -sub switch */
			error = TRUE;
		    else
		    {
			arg ++;
			*operation = SUBTRACTION;
		    } /* endif */
		}
                else if (strcmp(argv[arg], "-mult") == 0)
                {
                    if (gotmultfact || gotoperation)
                        /* already have a mult fact from */
                        /* a previous -m switch        */
                        error = TRUE;
                    else
                    {
                        gotmultfact = TRUE;
                        gotoperation = TRUE;
                        arg ++;
			context = MULT_FACT;
			*operation = MULTIPLICATION;
                    } /* endif */
                }
		else 
		    /* did not get recognizable switch */
		    error = TRUE;
		break;
	    case SNAP_FILES:
		/* processing all the snapshot files at once */

		/* counting all the snapshot files */
		for (i = arg; i < argc && *(argv[i]) != '-'; i ++)
		    *nsnapfilenames = *nsnapfilenames + 1;

		if (*nsnapfilenames <= 2)
		{
		    if (*snapfilenames = 
			(char **) malloc(*nsnapfilenames*sizeof(char *)))
		    {
			for (i = 0; arg+i < argc && *(argv[arg+i]) != '-'; i ++)
			    (*snapfilenames)[i] = argv[arg+i];

			if (arg+i < argc)
			    context = EXPECT_SWITCH;

			arg += i;
		    }
		    else
		    {
			fprintf(stderr, 
"ERROR: unable to malloc nfiles %d * sizeof(char**) = %d bytes for fnames\n", 
			    *nsnapfilenames, sizeof(char *), 
			    *nsnapfilenames*sizeof(char *));
			fflush(stderr);
			exit(1);
		    } /* endif */
		}
		else
		    /* too many snap file specified ... max is 2 */
		    error = TRUE;
		break;
	    case LOG_FILE:
		if (*(argv[arg]) == '-')
		    error = TRUE;
		else
		{
		    /* got logfilename */
		    *logfilename = argv[arg];
		    arg ++;
		    context = EXPECT_SWITCH;
		} /* endif */
		break;
            case COMMENTS:
		/* processing all the comments at once */

		/* counting the bytes in all the comments */
		for (i = arg, nbytes = 0; i < argc && *(argv[i]) != '-'; i ++)
		    nbytes = nbytes + strlen(argv[i]) + 1;

		if (nbytes)
		{
		    if (*comments = (char *) malloc(nbytes))
		    {
			**comments = '\0';

			for (i= arg; i < argc && *(argv[i]) != '-'; i ++)
			{
			    if (i != arg)
				strcat(*comments, " ");

			    strcat(*comments, argv[i]);
			} /* endfor */

			if (i < argc)
			{
			    context = EXPECT_SWITCH;
			    got_all_comments = TRUE;
			} /* endif */

			arg = i;
		    }
		    else
		    {
			fprintf(stderr, 
			    "ERROR: unable to malloc %d bytes for comments\n", 
			    nbytes);
			fflush(stderr);
			exit(1);
		    } /* endif */
		}
		else
		{
		    arg ++;
		    context = EXPECT_SWITCH;
		} /* endif */
                break;
            case KEYWORDS:
		/* processing all the keywords at once */

		/* counting the bytes in all the keywords */
		for (i = arg, nbytes = 0; i < argc && *(argv[i]) != '-'; i ++)
		    nbytes = nbytes + strlen(argv[i]) + 1;

		if (nbytes)
		{
		    if (*keywords = (char *) malloc(nbytes))
		    {
			**keywords = '\0';

			for (i= arg; i < argc && *(argv[i]) != '-'; i ++)
			{
			    if (i != arg)
				strcat(*keywords, " ");

			    strcat(*keywords, argv[i]);
			} /* endfor */

			if (i < argc)
			{
			    context = EXPECT_SWITCH;
			    got_all_keywords = TRUE;
			} /* endif */

			arg = i;
		    }
		    else
		    {
			fprintf(stderr, 
			    "ERROR: unable to malloc %d bytes for keywords\n", 
			    nbytes);
			fflush(stderr);
			exit(1);
		    } /* endif */
		}
		else
		{
		    arg ++;
		    context = EXPECT_SWITCH;
		} /* endif */
                break;
            case OUT_FILE:
                if (*(argv[arg]) == '-')
                    error = TRUE;
                else
                {
                    /* got outfilename */
                    *outfilename = argv[arg];
                    arg ++;
                    context = EXPECT_SWITCH;
                } /* endif */
                break;
	    case MULT_FACT:
		/* got multiplication factor */
		*mult_factor = strtod(argv[arg], &ret_cp);
		if (argv[arg] == ret_cp)
		    error = TRUE;
		else
		{
		    arg ++;
		    context = EXPECT_SWITCH;
		} /* endif */
		break;
	    default:
		error = TRUE;
		break;
	} /* end switch() */
    } /* endwhile */

/* printf("at end of loop error %c, nsnaps %d\n", (error ? 'T' : 'F'), *nsnapfilenames); */

    error |= (context == OUT_FILE || context == LOG_FILE 
		|| ((*operation == ADDITION || *operation == SUBTRACTION) 
			&& *nsnapfilenames != 2)
		|| (*operation == MULTIPLICATION && *nsnapfilenames != 1)
	    );

    if (error)
    {
printf("\n\tusage: %s (snap1 snap2 {-add or -sub}) or (snap1 -mult m)\n", argv[0]);
printf("\t\t{-l logfile} {-o outfile} {-v} {-c ... comments ...}\n");
printf("\t\t{-k keyword1 ... keywordn} {-sdds or -nosdds}\n\n");

printf("where\n\n");

printf("\tsnap1 snap2 {-add or -sub} - Adding/subtracting snapshots.  Here\n");
printf("\t\tyou wish to either add or subtract exactly two snapshot files.\n");
printf("\t\tYou must specify exactly two snapshot files when performing\n");
printf("\t\teither addition or subtraction.  The default operation is\n");
printf("\t\t-sub (subtraction).\n\n");

printf("\tsnap1 -mult m - Multiplying snapshots.  Here you wish to multiply\n");
printf("\t\tall the values in snap1 by multiplication factor m.\n\n");

printf("\t-l logfile - Log filename. The name of the file where all logging\n");
printf("\t\tmessages (e.g. error messages, reports of process variables\n");
printf("\t\tthat were not found) go.  The default is stderr.\n\n");

printf("\t-o outfile - Result filename.  The name of the file where the\n");
printf("\t\tresulting information goes.  The default is stdout.\n\n");

printf("\t-v - Verbose.  This increases the amount of information displayed\n");
printf("\t\tin the logfile.\n\n");

printf("\t-c ... comments ... - Comments.  Adds comments to the header of\n");
printf("\t\tthe result file.\n\n");

printf("\t-k keyword1 ... keywordn - Keywords.  Adds keywords to the header\n");
printf("\t\tof the result file.\n\n");

printf("\t-sdds or -nosdds - SDDS/non-SDDS snapshot file.  Explicitly\n");
printf("\t\tspecifying that the generated snapshot file will be \n");
printf("\t\tSDDS/non-SDDS compliant.  The default is to adopt the SDDS\n");
printf("\t\ttype from the input(s).  If there is a heterogeneous set of\n");
printf("\t\tinputs (some SDDS and some non-SDDS), the default is to produce\n");
printf("\t\tand SDDS compliant snapshot file.\n\n");

	exit(1);
    } /* endif */

} /* end parse_command_line_args() */

/****************************************************************
*
*
****************************************************************/

static void process_command_line_args(char *logfilename, char *outfilename, 
    char **snapfilenames, int nsnapsfilenames, Burtobjlistlist *snapshots, 
    char output_spec)
{

int i;
Burtobjlist *clp;
char input_profile;

    if (logfilename)
    {
	if (!(Log_fp = fopen(logfilename, "w")))
	{
	    printf("ERROR: unable to open logfile >%s<\n", logfilename);
	    exit(1);
	} /* endif */
    } 
    else
	Log_fp = stderr;

    if (Verbose)
    {
	fprintf(Log_fp, "**** ARITHMETIC BURT LOGFILE\n");
	fflush(Log_fp);
    } /* endif */

    if (!snapfilenames)
    {
        fprintf(Log_fp,
            "ERROR: process_command_line_args() got NULL snapfilenames\n");
        fflush(Log_fp);
        BurtExit(1);
    } /* endif */

    if (!snapshots)
    {
        fprintf(Log_fp,
            "ERROR: process_command_line_args() got NULL snapshots\n");
        fflush(Log_fp);
        BurtExit(1);
    } /* endif */

    if (Verbose)
    {
	fprintf(Log_fp, "\n--- Start processing snapshot files\n");
	fflush(Log_fp);
    } /* endif */

    for (i = 0; i < nsnapsfilenames; i ++)
    {
        if (BurtSDDSFile(snapfilenames[i]))
        {
            if (i == 0)
                input_profile = SDDSINPUT;
            else if (input_profile != SDDSINPUT)
                input_profile = HETROINPUT;
        }
        else
        {
            if (i == 0)
                input_profile = NONSDDSINPUT;
            else if (input_profile != NONSDDSINPUT)
                input_profile = HETROINPUT;
	} /* endif */

        if (clp = BurtProcessSnapshotFile(snapfilenames[i]))
            snapshots->append(clp);
        else
        {
            fprintf(Log_fp,
"ERROR: process_command_line_args() got NULL from BurtProcessSnapshotFile()\n");
            fflush(Log_fp);
            BurtExit(1);
        } /* endif */
    } /* endfor */

    if (Verbose)
    {
	fprintf(Log_fp, "--- End processing snapshot files\n");
	fflush(Log_fp);
    } /* endif */

    if (output_spec == REQSDDS)
        Sdds = TRUE;  /* explictly wants sdds */
    else if (output_spec == REQNOSDDS)
        Sdds = FALSE;  /* explictly does not want sdds */
    else if (input_profile == SDDSINPUT)
        Sdds = TRUE;  /* no specification, all input files sdds */
    else if (input_profile == NONSDDSINPUT)
        Sdds = FALSE;  /* no specification, all input files non-sdds */
    else
        Sdds = TRUE;  /* default ... no specifation, hetrogeneous input */

    if (!Sdds)
    {
	if (outfilename)
	{
	    if (!(Out_fp = fopen(outfilename, "w")))
	    {
		fprintf(Log_fp,
		    "ERROR: unable to open outfile >%s<\n", outfilename);
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */
	}
	else
	    Out_fp = stdout;
    } /* endif */

} /* end process_command_line_args() */

