/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#ifndef __burtobjs__
#define __burtobjs__

#include <string.h> /* for strdup() */
#include <stdlib.h> /* for free() */
#include <unistd.h> /* for uid_t and gid_t */
#include <sys/param.h> /* for MAXPATHLEN */


/* needed for PVNAME_SZ and FLDNAME_SZ */
#include <dbDefs.h>
/* for channel access calls */
#include <cadef.h>
#include <caerr.h>
#include <db_access.h>

#include "burtpublic.h"

#include "list.h"
#include "burtlist.h"

#include "SDDS.h"

#ifndef FLDNAME_SZ
#define FLDNAME_SZ 4  /*Field Name Size*/
#endif

#ifndef BOOL
#define BOOL char
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE 1
#endif

/* SnapParms types */
#define ABSOLUTE 0
#define RELATIVE 1
#define NOWRITE  2

/* Burtobj types */
#define PVAR      0
#define ATOMICDEV 1

#define MAXPVARNAMELENGTH ((PVNAME_SZ)+(FLDNAME_SZ)+2)

#define STATUS_DBR_REQ(d) ((d) == DBR_STS_STRING \
                        || (d) == DBR_STS_INT \
                        || (d) == DBR_STS_SHORT \
                        || (d) == DBR_STS_FLOAT \
                        || (d) == DBR_STS_ENUM \
                        || (d) == DBR_STS_CHAR \
                        || (d) == DBR_STS_LONG \
                        || (d) == DBR_STS_DOUBLE )

#define TIME_DBR_REQ(d) ((d) == DBR_TIME_STRING \
                        || (d) == DBR_TIME_INT \
                        || (d) == DBR_TIME_SHORT \
                        || (d) == DBR_TIME_FLOAT \
                        || (d) == DBR_TIME_ENUM \
                        || (d) == DBR_TIME_CHAR \
                        || (d) == DBR_TIME_LONG \
                        || (d) == DBR_TIME_DOUBLE )

#define GR_DBR_REQ(d) ((d) == DBR_GR_STRING \
                        || (d) == DBR_GR_INT \
                        || (d) == DBR_GR_SHORT \
                        || (d) == DBR_GR_FLOAT \
                        || (d) == DBR_GR_ENUM \
                        || (d) == DBR_GR_CHAR \
                        || (d) == DBR_GR_LONG \
                        || (d) == DBR_GR_DOUBLE )

#define CTRL_DBR_REQ(d) ((d) == DBR_CTRL_STRING \
                        || (d) == DBR_CTRL_INT \
                        || (d) == DBR_CTRL_SHORT \
                        || (d) == DBR_CTRL_FLOAT \
                        || (d) == DBR_CTRL_ENUM \
                        || (d) == DBR_CTRL_CHAR \
                        || (d) == DBR_CTRL_LONG \
                        || (d) == DBR_CTRL_DOUBLE )

#define VALID_ATOMIC_DBR_REQ(d) (VALID_DB_REQ((d)) \
			    && !STATUS_DBR_REQ((d)) \
			    && !TIME_DBR_REQ((d)) \
			    && !GR_DBR_REQ((d)) \
			    && !CTRL_DBR_REQ((d)))

/************/
/*          */
/* Filename */
/*          */
/************/

class Filename
{

public:
    Filename(const char *const dir, const char *const fname);
    Filename(const char *const fname);

    /* start inline methods */
    const char *get_dir() const;
    const char *get_fname() const;
    BOOL has_dir() const;
    /* end inline methods */

private:
    char directory[MAXPATHLEN+2];
    char filename[MAXPATHLEN+2];

}; /* end class Filename */

inline const char *Filename::get_dir() const { return directory; }
inline const char *Filename::get_fname() const { return filename; }
inline BOOL Filename::has_dir() const { return (*get_dir() ? TRUE : FALSE); }

/*********/
/*       */
/* Value */
/*       */
/*********/

class Value
{

public:
    Value();
    void clear_value();

    BOOL defined() const;

    void set(const char, const int, const void *const);

    void write_val(const char *const, const BOOL, const BOOL) const;
    void write_val_sdds(SDDS_TABLE *, const long &, const char *const, 
	const char *const, const char *const, const char *const, 
	const char *const, const BOOL, const BOOL) const;

    /* start inline methods */
    ~Value();
    char get_dbr_type() const;
    void *get_pval() const;
    int get_nelem() const;
    /* end inline methods */

private:
    /* THESE ARE NEVER CALLED ... JUST ELIMINATES ANNOYING COMPILER */
    /* WARNINGS ABOUT VARIABLES THAT CA DEFINES BUT NEVER USES ...  */
    /* COULD BE MEMBERS OF ANY CLASS ... Value ARBITRARILY CHOSEN.  */
    void never_called_but_eliminates_annoying_CA_warinigs1() const;
    void never_called_but_eliminates_annoying_CA_warinigs2() const;

    void set_dbr_type(const char);
    void set_nelem(const int);

    /* start inline methods */
    void set_pval(void *const);
    /* end inline methods */

    /* either all three are at their initial state which renders defined     */
    /* FALSE or all three are NOT at their initial state.  it is considered  */
    /* an error condition for some, but not all, of the three to be at their */
    /* initial state */

    char dbr_type;
    void *pval;
    int nelem;
}; /* end class Value */

inline void *Value::get_pval() const { return pval; }
inline Value::~Value() { free((char *) get_pval()); }
inline void Value::set_pval(void *const p) { pval = p; }
inline char Value::get_dbr_type() const { return dbr_type; }
inline int Value::get_nelem() const { return nelem; }

/***********/
/*         */
/* Channel */
/*         */
/***********/

class Channel
{

public:
    /* start inline methods */
    void clear_ever_searched();
    void set_ever_searched();

    BOOL ever_searched() const;
    /* NOTE: following two are NOT char *get_name() const; because cannot */
    /*       assure that the receiver of the pointer will not change this */
    /*       instance of the object.                                      */
    const char *get_name() const;
    chid *get_cid();

    void set_read_pending();
    void clear_read_pending();
    void set_reported();
    void clear_reported();

    int get_nreq() const;
    char get_dbr_req_type() const;
    BOOL read_pending() const;
    BOOL reported() const;

    Value *get_val() const;

    /* NOTE: following is NOT set_val(const Value *const) because although */
    /*       we promise not to change the value of the received pointer,   */
    /*       we do not promise to keep the thing pointed to a constant.    */

    void set_val(Value *const);

    /* end inline methods */

    Channel();

    BOOL connected() const;
    int channel_nelem() const;
    char channel_type() const;

    void set_name(const char *const);
    BOOL write_to_ioc();
    void set_dbr_req_type(const char);
    void set_nreq(const int);

private:
    char name[MAXPVARNAMELENGTH];
    chid cid;
    /* placed here because seg faults when do connected() on */
    /* a channel that has never been ca_search()'d           */
    BOOL ever_srchd;
    int nreq;
    char dbr_req_type;
    BOOL read_pend;
    BOOL rprtd;
    /* a pointer back to the Value in the container Burtobj */
    Value *vp;

}; /* end class Channel */


inline void Channel::set_ever_searched() { ever_srchd = TRUE; }
inline void Channel::clear_ever_searched() { ever_srchd = FALSE; }

inline BOOL Channel::ever_searched() const { return ever_srchd; }
inline const char *Channel::get_name() const { return (const char *) name; }
inline chid *Channel::get_cid() { return &cid; }

inline void Channel::set_read_pending() { read_pend = TRUE; }
inline void Channel::clear_read_pending() { read_pend = FALSE; }
inline void Channel::set_reported() { rprtd = TRUE; }
inline void Channel::clear_reported() { rprtd = FALSE; }

inline int Channel::get_nreq() const { return nreq; }
inline char Channel::get_dbr_req_type() const { return dbr_req_type; }
inline BOOL Channel::read_pending() const { return read_pend; }
inline BOOL Channel::reported() const { return rprtd; }

inline Value *Channel::get_val() const { return vp; }
inline void Channel::set_val(Value *const p) { vp = p; }

/**********/
/*        */
/* Sspvar */
/*        */
/**********/

class Sspvar
{

public:
    Sspvar();
    void set_print_type(const char);
    void set_name(const char *const);
    /* for debugging */
    void print() const;

    /* start inline methods */

    /* NOTE: following is NOT set_xxx(const char *const) because although  */
    /*       we promise not to change the value of the received pointer,   */
    /*       we do not promise to keep the thing pointed to a constant.    */

    void set_start(char *const);
    void set_end(char *const);
    void set_val(Value *const);

    char *get_start() const;
    char *get_end() const;

    const char *get_name() const;
    Value *get_val() const;
    char get_print_type() const;
    /* end inline methods */

private:
    char *start; /* points into Predicate::input_line */
    char *end;   /* points into Predicate::input_line */
    char name[MAXPVARNAMELENGTH]; /* holds name so later can be    */
				  /* searched for in some snapshot */
				  /* and populate Val *vp          */
    Value *vp;    /* to point to a Val in some snapshot */
    char print_type;

}; /* end class Sspvar */

inline void Sspvar::set_start(char *const c) { start = c; }
inline void Sspvar::set_end(char *const c) { end = c; }
inline void Sspvar::set_val(Value *const p) { vp = p; }

inline char *Sspvar::get_start() const { return start; }
inline char *Sspvar::get_end() const { return end; }
inline const char *Sspvar::get_name() const { return (const char *) name; }
inline Value *Sspvar::get_val() const { return vp; }
inline char Sspvar::get_print_type() const { return print_type; }

/*************/
/*           */
/* SnapParms */
/*           */
/*************/

/* These are the parameters associated with a single snapshot.  They are   */
/* placed into a list class called Snaplist.  Snaplist is a superclass     */
/* for any list class that is expected to hold a snapshot.                 */
/*                                                                         */
/* For this reason, there is a one-to-one correspondence between snapshots */
/* (single files) and a list, i.e., each snapshot gets its own list.       */

class SnapParms
{

public:
    SnapParms();
    ~SnapParms();

    void set_who_and_time_stamp();
    void set_currtime(const char *const);
    void set_loginid(const char *const);
    void set_real_world_name(const char *const);
    void set_comments(const char *const);
    void set_keywords(const char *const);
    void set_type(const char);

    void clear_snapparms();
    BOOL defined() const;

    /* start inline methods */
    char *get_currtime() const;
    char *get_loginid() const;
    char *get_real_world_name() const;
    uid_t get_effective_uid() const;
    gid_t get_group_id() const;
    char *get_comments() const;
    char *get_keywords() const;
    char get_type() const;

    void set_effective_uid(const uid_t &);
    void set_group_id(const gid_t &);

    BOOL is_absolute() const;
    BOOL is_relative() const;
    BOOL is_nowrite() const;

    void set_relative();
    void set_absolute();
    void set_nowrite();
    /* end inline methods */

private:
    /* either all vals are at their initial state which renders defined     */
    /* FALSE or all three are NOT at their initial state.  it is considered  */
    /* an error condition for some, but not all of them to be at their */
    /* initial state */

    char *currtime;
    char *loginid;
    char *real_world_name;
    uid_t effective_uid;
    gid_t group_id;
    char *comments;
    char *keywords;
    char type; /* ABSOLUTE, RELATIVE, or NOWRITE */

}; /* end class SnapParms */

inline char *SnapParms::get_currtime() const { return currtime; }
inline char *SnapParms::get_loginid() const { return loginid; }
inline char *SnapParms::get_real_world_name() const { return real_world_name; }
inline uid_t SnapParms::get_effective_uid() const { return effective_uid; }
inline gid_t SnapParms::get_group_id() const { return group_id; }
inline char *SnapParms::get_comments() const { return comments; }
inline char *SnapParms::get_keywords() const { return keywords; }
inline char SnapParms::get_type() const { return type; }

inline void SnapParms::set_effective_uid(const uid_t &e) { effective_uid = e; }
inline void SnapParms::set_group_id(const gid_t &g) { group_id = g; }

inline BOOL SnapParms::is_absolute() const { return (get_type() == ABSOLUTE); }
inline BOOL SnapParms::is_relative() const { return (get_type() == RELATIVE); }
inline BOOL SnapParms::is_nowrite() const { return (get_type() == NOWRITE); }

inline void SnapParms::set_relative() { set_type(RELATIVE); }
inline void SnapParms::set_absolute() { set_type(ABSOLUTE); }
inline void SnapParms::set_nowrite() { set_type(NOWRITE); }

/***********/
/*         */
/* Burtobj */
/*         */
/***********/

class Burtobj
{

public:

    /* start pure virtual ... must be provided by derived */
    virtual char isa() const = 0;
    virtual const char *get_name() const = 0;
    virtual const char *get_lineage() const = 0;
    virtual const char *get_readmsg() const = 0;
    virtual const char *get_writemsg() const = 0;
    virtual Channel *get_readchannel() = 0;
    virtual Channel *get_writechannel() = 0;
    /* end pure virtual ... must be provided by derived */

    /* start inline methods */
    Burtobj();
    virtual ~Burtobj () {};

    void set_read_only_notify();
    void clear_read_only_notify();

    void set_read_only();
    void clear_read_only();

    Value *get_val();
    const Value *get_val_RO() const;

    BOOL read_only_notify() const;
    BOOL read_only() const;
    /* end inline methods */

    void write_sdds(SDDS_TABLE *, const long &) const;

private:
    /* read only notify and read only kept here so we can read snapshots */
    /* into lists of these for predicates and diffs. */

    /* NOTE: about the relationship between read only notify and read only. */
    /*       RON=>RO.  That is, whenever read only notify is TRUE, so is    */
    /*       read only.  On the other hand, read only may be TRUE and read  */
    /*       only notify may be FALSE.  Read only notify is a more specific */
    /*       case of read only.                                             */

    BOOL RON;
    BOOL RO;
    Value val;

}; /* end class Burtobj */

inline void Burtobj::set_read_only_notify() { RON = TRUE; }
inline void Burtobj::clear_read_only_notify() { RON = FALSE; }

inline void Burtobj::set_read_only() { RO = TRUE; }
inline void Burtobj::clear_read_only() { RO = FALSE; }

inline Burtobj::Burtobj() { clear_read_only_notify(); clear_read_only();}

inline Value *Burtobj::get_val() { return &val; }
inline const Value *Burtobj::get_val_RO() const { return (const Value *) &val; }
inline BOOL Burtobj::read_only_notify() const { return RON; }
inline BOOL Burtobj::read_only() const { return (RO || RON); }

/************/
/*          */
/* Burtpvar */
/*          */
/************/

class Burtpvar : public Burtobj
{

public:
    /* start pure virtual ... must be provided because derived */
    char isa() const;
    const char *get_name() const;
    const char *get_lineage() const;
    const char *get_readmsg() const;
    const char *get_writemsg() const;
    Channel *get_readchannel();
    Channel *get_writechannel();
    /* end pure virtual ... must be provided because derived */

    /* start inline methods */
    Channel *get_channel();
    /* end inline methods */

private:
    Channel channel;

    /* start inline methods */
    const Channel *get_channel_RO() const;
    /* end inline methods */

}; /* end class Burtpvar */

inline char Burtpvar::isa() const { return PVAR; }
inline const Channel *Burtpvar::get_channel_RO() const { return (const Channel *) &channel; }
inline const char *Burtpvar::get_name() const { return get_channel_RO()->get_name(); }
inline const char *Burtpvar::get_lineage() const { return DEFAULTSTRING; }
inline const char *Burtpvar::get_readmsg() const { return DEFAULTSTRING; }
inline const char *Burtpvar::get_writemsg() const { return DEFAULTSTRING; }
inline Channel *Burtpvar::get_channel() { return &channel; }
inline Channel *Burtpvar::get_readchannel() { return  get_channel(); }
inline Channel *Burtpvar::get_writechannel() { return get_channel(); }


/***************/
/*             */
/* Burtatomdev */
/*             */
/***************/

class Burtatomdev : public Burtobj
{

public:
    /* start pure virtual ... must be provided because derived */
    char isa() const;
    const char *get_name() const;
    const char *get_lineage() const;
    const char *get_readmsg() const;
    const char *get_writemsg() const;
    Channel *get_readchannel();
    Channel *get_writechannel();
    /* end pure virtual ... must be provided because derived */

    Burtatomdev();
    void set_name(const char *const);
    void set_lineage(const char *const);
    void set_readmsg(const char *const);
    void set_writemsg(const char *const);

private:
    char name[MAXPVARNAMELENGTH];
    char lineage[MAXPVARNAMELENGTH];
    char readmsg[MAXPVARNAMELENGTH];
    char writemsg[MAXPVARNAMELENGTH];
    Channel readchannel;
    Channel writechannel;

}; /* end class Burtatomdev */

inline char Burtatomdev::isa() const { return ATOMICDEV; }
inline const char *Burtatomdev::get_name() const { return (const char *) name; }
inline const char *Burtatomdev::get_lineage() const { return (const char *) lineage; }
inline const char *Burtatomdev::get_readmsg() const { return (const char *) readmsg; }
inline const char *Burtatomdev::get_writemsg() const { return (const char *) writemsg; }
inline Channel *Burtatomdev::get_readchannel() { return &readchannel; }
inline Channel *Burtatomdev::get_writechannel() { return &writechannel; }

/***********/
/*         */
/* Iocpvar */
/*         */
/***********/

class Iocpvar
{

public:
    Iocpvar();
    /* for debugging */
    void print() const;

    /* start inline methods */
    void set_start(char *const);
    void set_end(char *const);
    char *get_start() const;
    char *get_end() const;
    Burtpvar *get_burtpvar() const;
    /* end inline methods */

private:
    char *start; /* points into Predicate::input_line */
    char *end;   /* points into Predicate::input_line */

    /* NOT the repsonsibility of Iocpvar to destroy this   */
    /* it is allocated during execution, but it is ALWAYS  */
    /* placed into a Predicatelist who owns it.  It is the */
    /* Predicatelist that's responsible for destroying it. */

    Burtpvar bpv;

}; /* end class Iocpvar */

inline void Iocpvar::set_start(char *const c) { start = c; }
inline void Iocpvar::set_end(char *const c) { end = c; }

inline char *Iocpvar::get_start() const { return start; }
inline char *Iocpvar::get_end() const { return end; }
inline Burtpvar *Iocpvar::get_burtpvar() const { return (Burtpvar *) &bpv; }

/*************/
/*           */
/* Predicate */
/*           */
/*************/

class Predicate
{

public:
    Predicate();
    ~Predicate();

    void set_raw_input_line(const char *const);
    void set_input_line(const char *const);
    void set_directive_on_fail(const char);
    /* for debugging */
    void print() const;

    /* start inline methods */
    List<Iocpvar> *get_ioc_pvars_list() const;
    List<Sspvar> *get_ss_pvars_list() const;
    char *get_raw_input_line() const;
    char *get_input_line() const;
    char *get_start() const;
    char get_directive_on_fail() const;

    void set_start(char *const);

    void clear_printed();
    void set_printed();
    BOOL printed() const;
    /* end inline methods */

private:
    /* both of these MUST NOT be owners */
    List<Iocpvar> ioc_pvars_list;
    List<Sspvar> ss_pvars_list;
    char *raw_input_line;
    char *input_line;
    char *start; /* pointer into input_line */
    char directive_on_fail;
    BOOL prnted;
}; /* end class Predicate */

inline List<Iocpvar> *Predicate::get_ioc_pvars_list() const { return (List<Iocpvar> *) &ioc_pvars_list; }
inline List<Sspvar> *Predicate::get_ss_pvars_list() const { return (List<Sspvar> *) &ss_pvars_list; }
inline char *Predicate::get_raw_input_line() const { return raw_input_line; }
inline char *Predicate::get_input_line() const { return input_line; }
inline char *Predicate::get_start() const { return start; }
inline char Predicate::get_directive_on_fail() const { return directive_on_fail; }
inline void Predicate::set_start(char *const cp) { start = cp; }
inline void Predicate::clear_printed() { prnted = FALSE; }
inline void Predicate::set_printed() { prnted = TRUE; }
inline BOOL Predicate::printed() const { return prnted; }

#endif
