/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#define MAINLINE

#include <stdlib.h>

#include "burtcommon.h"
#include "burtobjs.h"
#include "burtrb.h"

#include "list.h"
#include "burtlist.h"

/****************************************/
/*                                      */
/* Externally Visibile Remote Functions */
/*                                      */
/****************************************/

/* in burtcommon.cc */
void BurtCommonInit(const BOOL);
BOOL BurtPreprocessFile(const char *const, const char **const,
    const char **const, char *const);
void BurtProcessRequestFile(char *const, Burtobjlist *const);
BOOL BurtSDDSFile(const char *const);
void BurtExit(const int);

/***************************/
/*                         */
/* Private Local Functions */
/*                         */
/***************************/

static void init();
static void parse_command_line_args(int argc, char **argv, char **logfilename, 
    char **outfilename, char ***reqfilenames, int *nreqfilenames, 
    char **defines, char **includes, int *retry_count, BOOL *keep_generated, 
    char **comments, char **keywords, BOOL *verbose, char *output_spec);
static void process_command_line_args(char *logfilename, char *outfilename, 
    char **reqfilenames, int nreqfilenames, char **defines, char **includes,
    Burtobjlist *snapshot, BOOL keep_generated, char output_spec);

int main(int argc, char **argv)
{

char *logfilename;
char **reqfilenames;
int nreqfilenames;
char *defines[MAXDEFINES];
char *includes[MAXINCLUDES];
int retry_count;
char *comments;
char *keywords;
BOOL keep_generated;
Burtobjlist snapshot;
char output_spec;
BOOL found_all;

    init();
    snapshot.set_owner();

    parse_command_line_args(argc, argv, &logfilename, &Outfilename,
        &reqfilenames, &nreqfilenames, defines, includes, &retry_count, 
	&keep_generated, &comments, &keywords, &Verbose, &output_spec);

    process_command_line_args(logfilename, Outfilename, reqfilenames, 
	nreqfilenames, defines, includes, &snapshot, keep_generated, 
	output_spec);

    found_all = (snapshot.read_snapshot_from_iocs(retry_count, comments, 
		    keywords));
    snapshot.write_snapshot_to_file();

    if (reqfilenames)
	free((char *) reqfilenames);

    if (comments)
	free(comments);

    if (keywords)
	free(keywords);

    if (found_all)
	BurtExit(0);
    else
	BurtExit(1);

} /* end main() */

/******************************************************
*
*
******************************************************/

static void init()
{

    BurtCommonInit(TRUE);

} /* end init() */


/******************************************************
*
*
******************************************************/

static void parse_command_line_args(int argc, char **argv, char **logfilename, 
    char **outfilename, char ***reqfilenames, int *nreqfilenames, 
    char **defines, char **includes, int *retry_count, BOOL *keep_generated, 
    char **comments, char **keywords, BOOL *verbose, char *output_spec)
{

int arg;
int context;
int nbytes;
int nincludes;
int ndefines;
int i;
BOOL got_all_request_files;
BOOL got_all_comments;
BOOL got_all_keywords;
BOOL gotretrycount;
BOOL error;

    /* default values */
    *logfilename = (char *) NULL;
    *outfilename = (char *) NULL;
    *reqfilenames = (char **) NULL;
    *nreqfilenames = 0;
    *retry_count = 0;
    *keep_generated = FALSE;
    *comments = (char *) NULL;
    *keywords = (char *) NULL;
    *verbose = FALSE;
    *output_spec = NOREQ;
    for (ndefines = 0; ndefines < MAXDEFINES; ndefines ++)
	defines[ndefines] = (char *) NULL;
    for (nincludes = 0; nincludes < MAXINCLUDES; nincludes ++)
	includes[nincludes] = (char *) NULL;

    got_all_request_files = FALSE;
    gotretrycount = FALSE;
    got_all_comments = FALSE;
    got_all_keywords = FALSE;
    context = EXPECT_SWITCH;
    nincludes = ndefines = 0;
    arg = 1;
    error = FALSE;
    while (arg < argc && !error)
    {
/* printf("arg %d >%s< context %s\n", arg, argv[arg], (context == EXPECT_SWITCH ? "expect switch" : (context == REQ_FILES ? "request files" : (context == LOG_FILE ? "log file" : (context == RETRY_COUNT ? "retry count" : (context == COMMENTS ? "comments" : (context == KEYWORDS ? "keywords" : "unknown"))))))); */
	switch (context)
	{
	    case EXPECT_SWITCH:
		if (strncmp(argv[arg], "-D", 2) == 0)
		{
		    /* -D option */
		    if (ndefines < MAXDEFINES)
		    {
			defines[ndefines] = argv[arg];
			arg ++;
			ndefines ++;
		    }
		    else
			/* too many -D specifications */
			error = TRUE;
		}
		else if (strncmp(argv[arg], "-I", 2) == 0)
		{
		    /* -I option */
		    if (nincludes < MAXINCLUDES)
		    {
			includes[nincludes] = argv[arg];
			arg ++;
			nincludes ++;
		    }
		    else
			/* too many -I specifications */
			error = TRUE;
		}
		else if (strcmp(argv[arg], "-f") == 0)
		{
		    if (got_all_request_files)
			/* already got all the request files from */
			/* from a previous -f switch              */
			error = TRUE;
		    else
		    {
			arg ++;
			context = REQ_FILES;
		    } /* endif */
		}
		else if (strcmp(argv[arg], "-l") == 0)
		{
		    if (*logfilename)
			/* already have a logfile from */
			/* a previous -l switch        */
			error = TRUE;
		    else
		    {
			arg ++;
			context = LOG_FILE;
		    } /* endif */
		}
                else if (strcmp(argv[arg], "-c") == 0)
                {
                    if (got_all_comments)
                        /* already got all the comments from */
                        /* from a previous -c switch         */
                        error = TRUE;
                    else
                    {
                        arg ++;
                        context = COMMENTS;
                    } /* endif */
                }
                else if (strcmp(argv[arg], "-k") == 0)
                {
                    if (got_all_keywords)
                        /* already got all the keywords from */
                        /* from a previous -k switch         */
                        error = TRUE;
                    else
                    {
                        arg ++;
                        context = KEYWORDS;
                    } /* endif */
                }
		else if (strcmp(argv[arg], "-o") == 0)
		{
		    if (*outfilename)
			/* already have a outfile from */
			/* a previous -o switch        */
			error = TRUE;
		    else
		    {
			arg ++;
			context = OUT_FILE;
		    } /* endif */
		}
		else if (strcmp(argv[arg], "-r") == 0)
		{
		    if (gotretrycount)
			/* already have a seconds from */
			/* a previous -r switch        */
			error = TRUE;
		    else
		    {
			gotretrycount = TRUE;
			arg ++;
			context = RETRY_COUNT;
		    } /* endif */
		}
		else if (strcmp(argv[arg], "-d") == 0)
		{
		    arg ++;
		    *keep_generated = TRUE;
		}
		else if (strcmp(argv[arg], "-v") == 0)
		{
		    arg ++;
		    *verbose = TRUE;
		}
		else if (strcmp(argv[arg], "-sdds") == 0)
		{
		    if (*output_spec == NOREQ)
		    {
			arg ++;
			*output_spec = REQSDDS;
		    }
		    else
			error = TRUE;
		}
		else if (strcmp(argv[arg], "-nosdds") == 0)
		{
		    if (*output_spec == NOREQ)
		    {
			arg ++;
			*output_spec = REQNOSDDS;
		    }
		    else
			error = TRUE;
		}
		else 
		    /* did not get recognizable switch */
		    error = TRUE;
		break;
	    case REQ_FILES:
		/* processing all the request files at once */

		/* counting all the request files */
		for (i = arg; i < argc && *(argv[i]) != '-'; i ++)
		    *nreqfilenames = *nreqfilenames + 1;

		if (*nreqfilenames)
		{
		    if (*reqfilenames =
			(char **) malloc(*nreqfilenames*sizeof(char *)))
		    {
                        for (i = 0; arg+i < argc && *(argv[arg+i]) != '-'; i ++)
                            (*reqfilenames)[i] = argv[arg+i];

                        if (arg+i < argc)
			{
                            context = EXPECT_SWITCH;
			    got_all_request_files = TRUE;
			} /* endif */

                        arg += i;

		    }
		    else
		    {
                        fprintf(stderr,
"ERROR: unable to malloc nfiles %d * sizeof(char**) = %d bytes for fnames\n",
                            *nreqfilenames, sizeof(char *),
                            *nreqfilenames*sizeof(char *));
                        fflush(stderr);
                        exit(1);
		    } /* endif */
		}
		else
		{
		    arg ++;
		    context = EXPECT_SWITCH;
		} /* endif */
		break;
	    case LOG_FILE:
		if (*(argv[arg]) == '-')
		    error = TRUE;
		else
		{
		    /* got logfilename */
		    *logfilename = argv[arg];
		    arg ++;
		    context = EXPECT_SWITCH;
		} /* endif */
		break;
            case COMMENTS:
                /* processing all the comments at once */

                /* counting the bytes in all the comments */
                for (i = arg, nbytes = 0; i < argc && *(argv[i]) != '-'; i ++)
                    nbytes = nbytes + strlen(argv[i]) + 1;

		if (nbytes)
		{
		    if (*comments = (char *) malloc(nbytes))
		    {
			**comments = '\0';

			for (i= arg; i < argc && *(argv[i]) != '-'; i ++)
			{
			    if (i != arg)
				strcat(*comments, " ");

			    strcat(*comments, argv[i]);
			} /* endfor */

			if (i < argc)
			{
			    context = EXPECT_SWITCH;
			    got_all_comments = TRUE;
			} /* endif */

			arg = i;
		    }
		    else
		    {
			fprintf(stderr,
			    "ERROR: unable to malloc %d bytes for comments\n",
			    nbytes);
			fflush(stderr);
			exit(1);
		    } /* endif */
		}
		else
		{
		    arg ++;
		    context = EXPECT_SWITCH;
		} /* endif */
                break;
            case KEYWORDS:
                /* processing all the keywords at once */

                /* counting the bytes in all the keywords */
                for (i = arg, nbytes = 0; i < argc && *(argv[i]) != '-'; i ++)
                    nbytes = nbytes + strlen(argv[i]) + 1;

		if (nbytes)
		{
		    if (*keywords = (char *) malloc(nbytes))
		    {
			**keywords = '\0';

			for (i= arg; i < argc && *(argv[i]) != '-'; i ++)
			{
			    if (i != arg)
				strcat(*keywords, " ");

			    strcat(*keywords, argv[i]);
			} /* endfor */

			if (i < argc)
			{
			    context = EXPECT_SWITCH;
			    got_all_keywords = TRUE;
			} /* endif */

			arg = i;
		    }
		    else
		    {
			fprintf(stderr,
			    "ERROR: unable to malloc %d bytes for keywords\n",
			    nbytes);
			fflush(stderr);
			exit(1);
		    } /* endif */
		}
		else
		{
		    arg ++;
		    context = EXPECT_SWITCH;
		} /* endif */
                break;
	    case OUT_FILE:
		if (*(argv[arg]) == '-')
		    error = TRUE;
		else
		{
		    /* got outfilename */
		    *outfilename = argv[arg];
		    arg ++;
		    context = EXPECT_SWITCH;
		} /* endif */
		break;
	    case RETRY_COUNT:
		if (*(argv[arg]) == '-')
		    error = TRUE;
		else
		{
		    /* got wait seconds */
		    *retry_count = atoi(argv[arg]);
		    arg ++;
		    context = EXPECT_SWITCH;
		    if (*retry_count < 0)
			*retry_count = 0;
		} /* endif */
		break;
	    default:
		error = TRUE;
		break;
	} /* end switch() */
    } /* endwhile */

    error |= (context == RETRY_COUNT || context == LOG_FILE 
		|| (context == COMMENTS && !(*comments))
		|| (context == KEYWORDS && !(*keywords))
		|| context == OUT_FILE || !(*nreqfilenames));

    if (error)
    {
printf("\n\tusage: %s -f req1 {req2 ...} {-l logfile} {-o outfile} {-d} {-v}\n",
    argv[0]);
printf("\t\t{-c ... comments ...} {-k keyword1 ... keywordn}\n");
printf("\t\t{-r retry count} {-sdds or -nosdds} {-Dname{=def}}\n");
printf("\t\t{-Ipathname}\n\n");

printf("where\n\n");

printf("\t-f req1 {req2 ...} - Request filenames. This is the only switch\n");
printf("\t\tthat is not optional.  You must specify at least one request\n");
printf("\t\tfile.\n\n");

printf("\t-l logfile - Log filename. The name of the file where all logging\n");
printf("\t\tmessages (e.g. error messages, reports of process variables\n");
printf("\t\tthat were not found) go.  The default is stderr.\n\n");

printf("\t-o outfile - Snapshot filename.  The name of the file where the\n");
printf("\t\tsnapshot information goes.  The default is stdout.\n\n");

printf("\t-d - Debug.  Save the files created by processing the request\n");
printf("\t\tfiles with the C preprocessor.  The default is to delete these\n");
printf("\t\tfiles.\n\n");

printf("\t-v - Verbose.  This increases the amount of information displayed\n");
printf("\t\tin the logfile.\n\n");

printf("\t-c ... comments ... - Comments.  Adds comments to the header of\n");
printf("\t\tthe snapshot file.\n\n");

printf("\t-k keyword1 ... keywordn - Keywords.  Adds keywords to the header\n");
printf("\t\tof the snapshot file.\n\n");

printf("\t-r retry count - Number of additional attempts to wait for\n");
printf("\t\tconnections.  The program will attempt to find all the process\n");
printf("\t\tvariables.  If it is unsuccessful, it will try this many more\n");
printf("\t\ttimes to establish connections.  The default value is 0.\n\n");

printf("\t-sdds or -nosdds - SDDS/non-SDDS snapshot file.  Explicitly\n");
printf("\t\tspecifying that the generated snapshot file will be \n");
printf("\t\tSDDS/non-SDDS compliant.  The default is to adopt the SDDS\n");
printf("\t\ttype from the input(s).  If there is a heterogeneous set of\n");
printf("\t\tinputs (some SDDS and some non-SDDS), the default is to produce\n");
printf("\t\tand SDDS compliant snapshot file.\n\n");

printf("\t-Dname{=def} - Defines a symbol name to the C preprocessor.\n");
printf("\t\tEquivalent to a #define directive.  If no def is given, name\n");
printf("\t\tis defined as `1'.  The maximum number that can be specified\n");
printf("\t\tin this manner is %d.\n\n", MAXDEFINES);

printf("\t-Ipathname - Add pathname to the list of directories in which\n");
printf("\t\tto search for #include files with relative filenames (not\n");
printf("\t\tbeginning with slash `/').  The preprocessor first searches\n");
printf("\t\tfor #include files in the directory containing the request\n");
printf("\t\tfile, then in the directories named with -I option (if any).\n");
printf("\t\tThe maximumn number that can be specified in this manner\n");
printf("\t\tis %d.\n\n", MAXINCLUDES);

	exit(1);

    } /* endif */

} /* end parse_command_line_args() */

/******************************************************
*
*
******************************************************/

static void process_command_line_args(char *logfilename, char *outfilename, 
    char **reqfilenames, int nreqfilenames, char **defines, char **includes,
    Burtobjlist *snapshot, BOOL keep_generated, char output_spec)
{

int i, j;
char newfilename[MAXPATHLEN];
char input_profile;

    if (logfilename)
    {
	if (!(Log_fp = fopen(logfilename, "w")))
	{
	    printf("ERROR: unable to open logfile >%s<\n", logfilename);
	    exit(1);
	} /* endif */
    }
    else
	Log_fp = stderr;

    if (Verbose)
    {
	fprintf(Log_fp, "**** READ BURT LOGFILE\n");
	fflush(Log_fp);
    } /* endif */

    if (!reqfilenames)
    {
	fprintf(Log_fp,
	    "ERROR: process_command_line_args() got NULL reqfilenames\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!defines)
    {
	fprintf(Log_fp,
	    "ERROR: process_command_line_args() got NULL defines\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!includes)
    {
	fprintf(Log_fp,
	    "ERROR: process_command_line_args() got NULL includes\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!snapshot)
    {
	fprintf(Log_fp,
	    "ERROR: process_command_line_args() got NULL snapshot\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (Verbose)
    {
	fprintf(Log_fp, "\n--- Start processing files\n");
	fflush(Log_fp);
    } /* endif */

    for (i = 0; i < nreqfilenames; i ++)
    {
	if (Verbose)
	{
	    fprintf(Log_fp, "file >%s<\n", reqfilenames[i]);
	    fflush(Log_fp);
	} /* endif */

	snapshot->add_request_filename(reqfilenames[i]);

	if (BurtSDDSFile(reqfilenames[i]))
	{
	    if (i == 0)
		input_profile = SDDSINPUT;
	    else if (input_profile != SDDSINPUT)
		input_profile = HETROINPUT;

	    BurtProcessRequestFile(reqfilenames[i], snapshot);
	}
	else
	{
	    if (i == 0)
		input_profile = NONSDDSINPUT;
	    else if (input_profile != NONSDDSINPUT)
		input_profile = HETROINPUT;

	    if (!BurtPreprocessFile(reqfilenames[i], (const char**)defines, (const char**)includes, 
		newfilename))
	    {
		fprintf(Log_fp, 
"ERROR: process_command_line_args() could not preprocess non-SDDS file >%s<\n", 
		    reqfilenames[i]);
		fflush(Log_fp);

		BurtExit(1);
	    } /* endif */

	    BurtProcessRequestFile(newfilename, snapshot);

	    if (!keep_generated)
		unlink(newfilename);
	} /* endif */

    } /* endfor */

    if (Verbose)
    {
	fprintf(Log_fp, "--- End processing files\n");
	fflush(Log_fp);
    } /* endif */

    if (output_spec == REQSDDS)
	Sdds = TRUE;  /* explictly wants sdds */
    else if (output_spec == REQNOSDDS)
	Sdds = FALSE;  /* explictly does not want sdds */
    else if (input_profile == SDDSINPUT)
	Sdds = TRUE;  /* no specification, all input files sdds */
    else if (input_profile == NONSDDSINPUT)
	Sdds = FALSE;  /* no specification, all input files non-sdds */
    else
	Sdds = TRUE;  /* default ... no specifation, hetrogeneous input */

    if (Sdds)
    {
	/* cannot have duplicate request files if SDDS because cannot */
	/* have duplicate associate filenames in a single output file */

	/* checking that all the request filenames are unique */
	for (i = 1; i < nreqfilenames; i ++)
	{
	    for (j = i - 1; j >= 0; j --)
	    {
		if (!strcmp(reqfilenames[i], reqfilenames[j]))
		{
		    /* two that are the same */
		    fprintf(Log_fp, 
"ERROR: cannot have duplicate req filename >%s< when generating SDDS snapshot file\n", 
			reqfilenames[i]);
		    fflush(Log_fp);
		    BurtExit(1);
		} /* endif */
	    } /* endfor */
	} /* endfor */
    }
    else
    {
	if (outfilename)
	{
	    if (!(Out_fp = fopen(outfilename, "w")))
	    {
		fprintf(Log_fp,
		    "ERROR: unable to open outfile >%s<\n", outfilename);
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */
	}
	else
	    Out_fp = stdout;
    } /* endif */

} /* end process_command_line_args() */

