/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#ifndef __list__
#define __list__

#include <stdio.h> /* for NULL */

class Node;
template <class T> class Cursor;

/* These are ORDERED lists and duplicates are allowed ... meaning when   */
/* removing an item from the list, it removes the FIRST occurrence only. */

/********/
/*      */
/* List */
/*      */
/********/

template <class T>
class List
{

public:
    List();
    ~List();

    void set_owner();
    void set_view();
    char empty() const;
    char is_owner() const;
    int nelem() const;

    void prepend(const T *const);
    void append(const T *const);

    T *operator[](const int) const;

    T *remove(const T *const);

    Node *gethead() const; 
    Node *gettail() const;

private:
    Node *head; 
    Node *tail;
    char owner; /* TRUE=>owner, FALSE=>view */
		/* if owner, then when list is */
		/* destroyed it not only deletes */
		/* its nodes, but the objects also */
};

class Node
{

public:
    Node(const void *const);
    const void *thisone() const;

    void setprev(Node *const);
    Node *getprev() const;

    void setnext(Node *const);
    Node *getnext() const;

private:
    class Node *prev;
    class Node *next;

    const void *mine;

}; /* end class Node */

inline Node::Node(const void *const p) { prev = next = (Node *) NULL; mine = p;}
inline const void *Node::thisone() const { return mine; }

inline Node *Node::getprev() const { return prev; }
inline void Node::setprev(Node *const np) { prev = np; }

inline Node *Node::getnext() const { return next; }
inline void Node::setnext(Node *const np) { next = np; }

template <class T>
class Cursor
{

public:
    Cursor(List<T> *const);

    T *getfirst();
    T *getnext();

    T *getlast();
    T *getprev();

private:
    List<T> *mylist;
    Node *curr;

}; /* end class ClassOnecursor */



/******************/
/* Public Methods */
/******************/

/****************************************************************
*
*
****************************************************************/

template <class T>
List<T>::List()
{
    set_view();
    head = (Node *) NULL;
    tail = (Node *) NULL;

}

/****************************************************************
*
*
****************************************************************/

template <class T>
List<T>::~List()
{

Node *np;

    while (head)
    {
	np = head;
	head = np->getnext();

	if (is_owner())
	    delete (T *) (np->thisone());
	delete np;

    } /* endwhile */
}

/****************************************************************
*
*
****************************************************************/

template <class T>
void List<T>::set_owner() { owner = 1; }
template <class T>
void List<T>::set_view() { owner = 0; }

template <class T>
char List<T>::empty() const { return (head ? 0 : 1); }
template <class T>
char List<T>::is_owner() const { return owner; }


/****************************************************************
*
*
****************************************************************/

template <class T>
int List<T>::nelem() const
{

int rc;
Node *np;

    for (rc = 0, np = gethead(); np; np = np->getnext(), rc ++)
    ;

    return rc;
}

/****************************************************************
*
*
****************************************************************/

template <class T>
void List<T>::prepend(const T *const p)
{

Node *np;

    if (p)
    {
	np = new Node((void *const) p);

	np->setnext(head);

	if (head)
	    head->setprev(np);

	head = np;

	if (!tail)
	    tail = head;
    } /* endif */

} 

/****************************************************************
*
*
****************************************************************/

template <class T>
void List<T>::append(const T *const p)
{

Node *np;

    if (p)
    {
	np = new Node((void *const) p);

	np->setprev(tail);

	if (tail)
	    tail->setnext(np);

	tail = np;

	if (!head)
	    head = tail;
    } /* endif */

} 

/****************************************************************
*
*
****************************************************************/

template <class T>
T *List<T>::remove(const T *const p)
{

Node *np;
T *rc;

    if (p)
    {
	for (np = head; 
	    np && ((T *) (np->getnext())->thisone()) != p; 
		np = np->getnext())
	;

	if (np)
	{
	    /* found our guy and np points to him */
	    if (np->getprev())
		/* not first in list */
		(np->getprev())->setnext(np->getnext());
	    else
		/* first in list */
		head = np->getnext();

	    if (np->getnext())
		/* not last in list */
		(np->getnext())->setprev(np->getprev());
	    else
		/* last in list */
		tail = np->getprev();
	}
	else
	    /* not in list */
	    rc = (T *) NULL;
    } 
    else
	rc = (T *) NULL;

    return rc;

}

/****************************************************************
*
*
****************************************************************/

template <class T>
T *List<T>::operator[](const int index) const
{

Node *np;
int i;

    for(i = 0, np = gethead(); np && i < index; np = np->getnext(), i++)
    ;

    return (np ? (T *) np->thisone() : (T *) NULL);

}

/*******************/
/* Private Methods */
/*******************/

template <class T>
Node *List<T>::gethead() const { return head; }
template <class T>
Node *List<T>::gettail() const { return tail; }

/**********/
/*        */
/* Cursor */
/*        */
/**********/

/****************************************************************
*
*
****************************************************************/

template <class T>
Cursor<T>::Cursor(List<T> *const lp)
{
    mylist = lp;
    curr = (Node *) NULL;
}

/****************************************************************
*
*
****************************************************************/

template <class T>
T *Cursor<T>::getfirst() 
{
    return ((curr = mylist->gethead()) 
		? (T *) curr->thisone() 
		: (T *) NULL);
}

/****************************************************************
*
*
****************************************************************/

template <class T>
T *Cursor<T>::getnext() 
{

T *rc;

    if (curr)
	if (curr = curr->getnext())
	    rc = (T *) curr->thisone();
	else
	    rc = (T *) NULL;
    else
	rc = (T *) NULL;

    return rc;

}

/****************************************************************
*
*
****************************************************************/

template <class T>
T *Cursor<T>::getlast() 
{
    return ((curr = mylist->gettail()) 
		? (T *) curr->thisone() 
		: (T *) NULL);
}

/****************************************************************
*
*
****************************************************************/

template <class T>
T *Cursor<T>::getprev() 
{

T *rc;

    if (curr)
	if (curr = curr->getprev())
	    rc = (T *) curr->thisone();
	else
	    rc = (T *) NULL;
    else
	rc = (T *) NULL;

    return rc;

}


#endif
